from unittest import TestCase

from pyinfra.api import Config, Inventory, State
from pyinfra.api.arguments import pop_global_arguments


class TestOperationKwargs(TestCase):
    def test_get_from_config(self):
        config = Config(SUDO='config-value')
        inventory = Inventory((('somehost',), {}))

        state = State(config=config, inventory=inventory)

        kwargs, keys = pop_global_arguments(state, inventory.get_host('somehost'), {})
        assert kwargs['sudo'] == 'config-value'

    def test_get_from_host(self):
        config = Config(SUDO='config-value')
        inventory = Inventory(([('somehost', {'sudo': 'host-value'})], {}))

        state = State(config=config, inventory=inventory)

        kwargs, keys = pop_global_arguments(state, inventory.get_host('somehost'), {})
        assert kwargs['sudo'] == 'host-value'

    def test_get_from_state_deploy_kwargs(self):
        config = Config(SUDO='config-value')
        inventory = Inventory(([('somehost', {'sudo': 'host-value'})], {}))
        somehost = inventory.get_host('somehost')

        state = State(config=config, inventory=inventory)
        somehost.current_deploy_kwargs = {'sudo': 'deploy-kwarg-value'}

        kwargs, keys = pop_global_arguments(state, somehost, {})
        assert kwargs['sudo'] == 'deploy-kwarg-value'

    def test_get_from_kwargs(self):
        config = Config(SUDO='config-value')
        inventory = Inventory(([('somehost', {'sudo': 'host-value'})], {}))
        somehost = inventory.get_host('somehost')

        state = State(config=config, inventory=inventory)
        somehost.current_deploy_kwargs = {
            'sudo': 'deploy-kwarg-value',
            'sudo_user': 'deploy-kwarg-user',
        }

        kwargs, keys = pop_global_arguments(
            state,
            somehost,
            {'sudo': 'kwarg-value'},
        )
        assert kwargs['sudo'] == 'kwarg-value'
        assert kwargs['sudo_user'] == 'deploy-kwarg-user'
        assert 'sudo' in keys



import re

from pyinfra.api import FactBase


# Valid unit names consist of a "name prefix" and a dot and a suffix specifying the unit type.
# The "unit prefix" must consist of one or more valid characters
# (ASCII letters, digits, ":", "-", "_", ".", and "\").
# The total length of the unit name including the suffix must not exceed 256 characters.
# The type suffix must be one of
# ".service", ".socket", ".device", ".mount", ".automount",
# ".swap", ".target", ".path", ".timer", ".slice", or ".scope".
# Units names can be parameterized by a single argument called the "instance name".
# A template unit must have a single "@" at the end of the name (right before the type suffix).
# The name of the full unit is formed by inserting the instance name
# between "@" and the unit type suffix.
SYSTEMD_UNIT_NAME_REGEX = (
    r'[a-zA-Z0-9\:\-\_\.\\\@]+\.'
    r'(?:service|socket|device|mount|automount|swap|target|path|timer|slice|scope)'
)


def _make_systemctl_cmd(user_mode=False, machine=None, user_name=None):
    # base command for normal and user mode
    systemctl_cmd = 'systemctl --user' if user_mode else 'systemctl'

    # add user and machine flag if given in args
    if machine is not None:
        if user_name is not None:
            machine_opt = '--machine={1}@{0}'.format(machine, user_name)
        else:
            machine_opt = '--machine={0}'.format(machine)

        systemctl_cmd = '{0} {1}'.format(systemctl_cmd, machine_opt)

    return systemctl_cmd


class SystemdStatus(FactBase):
    '''
    Returns a dict of name -> status for systemd managed services.
    '''

    requires_command = 'systemctl'

    default = dict

    regex = r'^({systemd_unit_name_regex})\s+[a-z\-]+\s+[a-z]+\s+([a-z]+)'.format(
        systemd_unit_name_regex=SYSTEMD_UNIT_NAME_REGEX)

    def command(self, user_mode=False, machine=None, user_name=None):
        fact_cmd = _make_systemctl_cmd(
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        )

        return '{0} -al --plain --no-legend list-units'.format(fact_cmd)

    def process(self, output):
        services = {}

        for line in output:
            line = line.strip()
            matches = re.match(self.regex, line)
            if matches:
                is_active = matches.group(2) in ('running', 'waiting', 'exited')
                services[matches.group(1)] = is_active

        return services


class SystemdEnabled(FactBase):
    '''
    Returns a dict of name -> whether enabled for systemd managed services.
    '''

    requires_command = 'systemctl'

    default = dict

    regex = r'^({systemd_unit_name_regex})\s+([a-z]+)'.format(
        systemd_unit_name_regex=SYSTEMD_UNIT_NAME_REGEX)

    def command(self, user_mode=False, machine=None, user_name=None):
        fact_cmd = _make_systemctl_cmd(
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        )

        return (
            '{0} -al --plain --no-legend --state=loaded list-units | '
            'while read -r UNIT REST; do '
            'STATE=$({0} -P UnitFileState show -- "$UNIT"); '
            'if [ -n "$STATE" ]; then '
            'echo "$UNIT" "$STATE"; '
            'fi; '
            'done'
        ).format(fact_cmd)

    def process(self, output):
        units = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                units[matches.group(1)] = (
                    matches.group(2) in ('enabled', 'static')
                )

        return units



from os import path
from unittest import TestCase

from pyinfra.context import ctx_state
from pyinfra_cli.main import _main

from .util import run_cli
from ..paramiko_util import PatchSSHTestCase


class TestCliEagerFlags(TestCase):
    def test_print_help(self):
        result = run_cli('--version')
        assert result.exit_code == 0, result.stdout

        result = run_cli('--help')
        assert result.exit_code == 0, result.stdout


class TestDeployCli(PatchSSHTestCase):
    def setUp(self):
        ctx_state.reset()

    def test_invalid_deploy(self):
        result = run_cli(
            '@local',
            'not-a-file.py',
        )
        assert result.exit_code == 1, result.stdout
        assert 'No deploy file: not-a-file.py' in result.stdout


class TestOperationCli(PatchSSHTestCase):
    def test_invalid_operation_module(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'not_a_module.shell',
        )
        assert result.exit_code == 1, result.stdout
        assert 'No such module: not_a_module'

    def test_invalid_operation_function(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'server.not_an_operation',
        )
        assert result.exit_code == 1, result.stdout
        assert 'No such operation: server.not_an_operation'

    def test_deploy_operation(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'server.shell',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_deploy_operation_with_all(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventory_all.py'),
            'server.shell',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_deploy_operation_json_args(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventory_all.py'),
            'server.shell',
            '[["echo hi"], {}]',
        )
        assert result.exit_code == 0, result.stdout


class TestFactCli(PatchSSHTestCase):
    def test_get_fact(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'fact',
            'server.Os',
        )
        assert result.exit_code == 0, result.stdout
        assert '"somehost": null' in result.stdout

    def test_get_fact_with_kwargs(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'fact',
            'files.File',
            'path=.',
        )
        assert result.exit_code == 0, result.stdout
        assert '"somehost": null' in result.stdout

    def test_invalid_fact_module(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'fact',
            'not_a_module.NotAFact',
        )
        assert result.exit_code == 1, result.stdout
        assert 'No such module: `not_a_module`' in result.stdout

    def test_invalid_fact_class(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'fact',
            'server.NotAFact',
        )
        assert result.exit_code == 1, result.stdout
        assert 'No such fact: `server.NotAFact`' in result.stdout


class TestExecCli(PatchSSHTestCase):
    def test_exec_command(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_options(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--sudo',
            '--sudo-user', 'pyinfra',
            '--su-user', 'pyinfrawhat',
            '--port', '1022',
            '--user', 'ubuntu',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_serial(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--serial',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_no_wait(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--no-wait',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_debug_operations(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--debug-operations',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_debug_facts(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--debug-facts',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stdout


class TestDirectMainExecution(PatchSSHTestCase):
    '''
    These tests are very similar as above, without the click wrappers - basically
    here because coverage.py fails to properly detect all the code under the wrapper.
    '''

    def test_deploy_operation_direct(self):
        with self.assertRaises(SystemExit) as e:
            _main(
                inventory=path.join('tests', 'test_deploy', 'inventories', 'inventory.py'),
                operations=['server.shell', 'echo hi'],
                chdir=None,
                group_data=None,
                verbosity=0, ssh_user=None, ssh_port=None, ssh_key=None, ssh_key_password=None,
                ssh_password=None,
                sudo=False, sudo_user=None, use_sudo_password=False, su_user=None,
                parallel=None, fail_percent=0, dry=False, limit=None, no_wait=False, serial=False,
                winrm_username=None, winrm_password=None, winrm_port=None, winrm_transport=None,
                shell_executable=None, quiet=False, data=tuple(),
                debug=False, debug_facts=False, debug_operations=False,
                config_filename='config.py',
            )
            assert e.args == (0,)



from contextlib import contextmanager

import click
from gevent.lock import BoundedSemaphore

from pyinfra import logger
from pyinfra.connectors import EXECUTION_CONNECTORS
from pyinfra.connectors.util import remove_any_sudo_askpass_file

from .exceptions import ConnectError
from .facts import create_host_fact, delete_host_fact, get_host_fact


def extract_callable_datas(datas):
    for data in datas:
        # Support for dynamic data, ie @deploy wrapped data defaults where
        # the data is stored on the state temporarily.
        if callable(data):
            data = data()

        yield data


class HostData(object):
    '''
    Combines multiple AttrData's to search for attributes.
    '''

    override_datas = None

    def __init__(self, host, *datas):
        self.__dict__['host'] = host

        datas = list(datas)

        # Inject an empty override data so we can assign during deploy
        self.__dict__['override_datas'] = {}
        datas.insert(0, self.override_datas)

        self.__dict__['datas'] = tuple(datas)

    def __getattr__(self, key):
        for data in extract_callable_datas(self.datas):
            try:
                return data[key]
            except KeyError:
                pass

        raise AttributeError(f'Host `{self.host}` has no data `{key}`')

    def __setattr__(self, key, value):
        self.override_datas[key] = value

    def __str__(self):
        return str(self.datas)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def dict(self):
        out = {}

        # Copy and reverse data objects (such that the first items override
        # the last, matching __getattr__ output).
        datas = list(self.datas)
        datas.reverse()

        for data in extract_callable_datas(datas):
            out.update(data)

        return out


class Host(object):
    '''
    Represents a target host. Thin class that links up to facts and host/group
    data.
    '''

    connection = None
    state = None

    # Current context inside an @operation function
    in_op = False
    current_op_hash = None
    current_op_global_kwargs = None

    # Current context inside a @deploy function
    in_deploy = False
    current_deploy_name = None
    current_deploy_kwargs = None
    current_deploy_data = None
    current_deploy_op_order = None

    def __init__(
        self, name, inventory, groups,
        executor=EXECUTION_CONNECTORS['ssh'],
    ):
        self.inventory = inventory
        self.groups = groups
        self.executor = executor
        self.name = name

        # Arbitrary dict for connector use
        self.connector_data = {}

        # Fact data store
        self.facts = {}
        self.facts_lock = BoundedSemaphore()

        # Create the (waterfall data: override, host, group, global)
        self.data = HostData(
            self,
            lambda: inventory.get_override_data(),
            lambda: inventory.get_host_data(name),
            lambda: inventory.get_groups_data(groups),
            lambda: inventory.get_data(),
            # @deploy function data are default values, so come last
            self.get_deploy_data,
        )

    def __str__(self):
        return '{0}'.format(self.name)

    def __repr__(self):
        return 'Host({0})'.format(self.name)

    @property
    def connected(self):
        return self.connection is not None

    @property
    def host_data(self):
        return self.inventory.get_host_data(self.name)

    @property
    def group_data(self):
        return self.inventory.get_groups_data(self.groups)

    @property
    def print_prefix(self):
        return '{0}[{1}] '.format(
            click.style(''),  # reset
            click.style(self.name, bold=True),
        )

    def style_print_prefix(self, *args, **kwargs):
        return '{0}[{1}] '.format(
            click.style(''),  # reset
            click.style(self.name, *args, **kwargs),
        )

    def get_deploy_data(self):
        if self.current_deploy_data:
            return self.current_deploy_data

        return {}

    def noop(self, description):
        '''
        Log a description for a noop operation.
        '''

        handler = logger.info if self.state.print_noop_info else logger.debug
        handler('{0}noop: {1}'.format(self.print_prefix, description))

    @contextmanager
    def deploy(self, name, kwargs, data, in_deploy=True, deploy_op_order=None):
        '''
        Wraps a group of operations as a deploy, this should not be used
        directly, instead use ``pyinfra.api.deploy.deploy``.
        '''

        # Handle nested deploy names
        if self.current_deploy_name:
            name = '{0} | {1}'.format(self.current_deploy_name, name)

        # Store the previous values
        old_in_deploy = self.in_deploy
        old_deploy_name = self.current_deploy_name
        old_deploy_kwargs = self.current_deploy_kwargs
        old_deploy_data = self.current_deploy_data
        old_deploy_op_order = self.current_deploy_op_order
        self.in_deploy = in_deploy

        # Set the new values
        self.current_deploy_name = name
        self.current_deploy_kwargs = kwargs
        self.current_deploy_data = data
        self.current_deploy_op_order = deploy_op_order
        logger.debug('Starting deploy {0} (args={1}, data={2})'.format(
            name, kwargs, data,
        ))

        yield

        # Restore the previous values
        self.in_deploy = old_in_deploy
        self.current_deploy_name = old_deploy_name
        self.current_deploy_kwargs = old_deploy_kwargs
        self.current_deploy_data = old_deploy_data
        self.current_deploy_op_order = old_deploy_op_order

        logger.debug('Reset deploy to {0} (args={1}, data={2})'.format(
            old_deploy_name, old_deploy_kwargs, old_deploy_data,
        ))

    # Host facts
    #

    def get_fact(self, name_or_cls, *args, **kwargs):
        return get_host_fact(self.state, self, name_or_cls, args=args, kwargs=kwargs)

    def create_fact(self, name_or_cls, data=None, kwargs=None):
        return create_host_fact(self.state, self, name_or_cls, data, kwargs=kwargs)

    def delete_fact(self, name_or_cls, kwargs=None):
        return delete_host_fact(self.state, self, name_or_cls, kwargs=kwargs)

    # Connector proxy
    #

    def _check_state(self):
        if not self.state:
            raise TypeError('Cannot call this function with no state!')

    def connect(self, reason=None, show_errors=True):
        self._check_state()
        if not self.connection:
            self.state.trigger_callbacks('host_before_connect', self)

            try:
                self.connection = self.executor.connect(self.state, self)
            except ConnectError as e:
                if show_errors:
                    log_message = '{0}{1}'.format(
                        self.print_prefix,
                        click.style(e.args[0], 'red'),
                    )
                    logger.error(log_message)

                self.state.trigger_callbacks('host_connect_error', self, e)
            else:
                log_message = '{0}{1}'.format(
                    self.print_prefix,
                    click.style('Connected', 'green'),
                )
                if reason:
                    log_message = '{0}{1}'.format(
                        log_message,
                        ' ({0})'.format(reason),
                    )

                logger.info(log_message)
                self.state.trigger_callbacks('host_connect', self)

        return self.connection

    def disconnect(self):
        self._check_state()

        # Disconnect is an optional function for executors if needed
        disconnect_func = getattr(self.executor, 'disconnect', None)
        if disconnect_func:
            return disconnect_func(self.state, self)

        # TODO: consider whether this should be here!
        remove_any_sudo_askpass_file(self)

        self.state.trigger_callbacks('host_disconnect', self)

    def run_shell_command(self, *args, **kwargs):
        self._check_state()
        return self.executor.run_shell_command(self.state, self, *args, **kwargs)

    def put_file(self, *args, **kwargs):
        self._check_state()
        return self.executor.put_file(self.state, self, *args, **kwargs)

    def get_file(self, *args, **kwargs):
        self._check_state()
        return self.executor.get_file(self.state, self, *args, **kwargs)

    # Rsync - optional executor specific ability

    def check_can_rsync(self):
        check_can_rsync_func = getattr(self.executor, 'check_can_rsync', None)
        if check_can_rsync_func:
            return check_can_rsync_func(self)

        raise NotImplementedError('The {0} connector does not support rsync!'.format(
            self.executor.__name__,
        ))

    def rsync(self, *args, **kwargs):
        self._check_state()
        return self.executor.rsync(self.state, self, *args, **kwargs)



'''
Deploys come in two forms: on-disk, eg deploy.py, and @deploy wrapped functions.
The latter enable re-usable (across CLI and API based execution) pyinfra extension
creation (eg pyinfra-openstack).
'''

from functools import wraps

import pyinfra

from pyinfra import host, logger, state
from pyinfra.context import ctx_host, ctx_state

from .arguments import pop_global_arguments
from .exceptions import PyinfraError
from .host import Host
from .util import get_args_kwargs_spec, get_call_location, get_caller_frameinfo, memoize


@memoize
def show_state_host_arguments_warning(call_location):
    logger.warning((
        '{0}:\n\tLegacy deploy function detected! Deploys should no longer define '
        '`state` and `host` arguments.'
    ).format(call_location))


def add_deploy(state, deploy_func, *args, **kwargs):
    '''
    Prepare & add an deploy to pyinfra.state by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        deploy_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    '''

    if pyinfra.is_cli:
        raise PyinfraError((
            '`add_deploy` should not be called when pyinfra is executing in CLI mode! ({0})'
        ).format(get_call_location()))

    hosts = kwargs.pop('host', state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    # Append operations called in this deploy to the current order
    kwargs['_op_order_number'] = len(state.op_meta)

    with ctx_state.use(state):
        for deploy_host in hosts:
            with ctx_host.use(deploy_host):
                deploy_func(*args, **kwargs)


def deploy(func_or_name, data_defaults=None, _call_location=None):
    '''
    Decorator that takes a deploy function (normally from a pyinfra_* package)
    and wraps any operations called inside with any deploy-wide kwargs/data.
    '''

    # If not decorating, return function with config attached
    if isinstance(func_or_name, str):
        name = func_or_name

        def decorator(f):
            f.deploy_name = name
            if data_defaults:
                f.deploy_data = data_defaults
            return deploy(f, _call_location=get_call_location())
        return decorator

    # Actually decorate!
    func = func_or_name

    # Check whether an operation is "legacy" - ie contains state=None, host=None kwargs
    # TODO: remove this in v3
    is_legacy = False
    args, kwargs = get_args_kwargs_spec(func)
    if all(key in kwargs and kwargs[key] is None for key in ('state', 'host')):
        show_state_host_arguments_warning(_call_location or get_call_location())
        is_legacy = True
    func.is_legacy = is_legacy

    @wraps(func)
    def decorated_func(*args, **kwargs):
        deploy_kwargs, _ = pop_global_arguments(state, host, kwargs)

        # If this is a legacy operation function (ie - state & host arg kwargs), ensure that state
        # and host are included as kwargs.
        if func.is_legacy:
            if 'state' not in kwargs:
                kwargs['state'] = state
            if 'host' not in kwargs:
                kwargs['host'] = host
        # If not legacy, pop off any state/host kwargs that may come from legacy @deploy functions
        else:
            kwargs.pop('state', None)
            kwargs.pop('host', None)

        # Name the deploy
        deploy_name = getattr(func, 'deploy_name', func.__name__)
        deploy_data = getattr(func, 'deploy_data', None)

        deploy_op_order = None

        if not pyinfra.is_cli:
            # API mode: `add_deploy` provides the order number
            op_order_number = kwargs.pop('_op_order_number', None)
            if op_order_number is not None:
                deploy_op_order = [op_order_number]
            # API mode: nested deploy wrapped function call
            elif host.current_deploy_op_order:
                frameinfo = get_caller_frameinfo()
                deploy_op_order = host.current_deploy_op_order + [frameinfo.lineno]
            else:
                raise PyinfraError((
                    'Operation order number not provided in API mode - '
                    'you must use `add_deploy` to add operations.'
                ))

        with host.deploy(
            name=deploy_name,
            kwargs=deploy_kwargs,
            data=deploy_data,
            deploy_op_order=deploy_op_order,
        ):
            return func(*args, **kwargs)

    return decorated_func



import logging
import sys
import warnings

from fnmatch import fnmatch
from os import chdir as os_chdir, getcwd, path

import click

from pyinfra import (
    __version__,
    logger,
    state,
)
from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import NoGroupError, PyinfraError
from pyinfra.api.facts import get_facts
from pyinfra.api.operation import add_op
from pyinfra.api.operations import run_ops
from pyinfra.api.util import get_kwargs_str
from pyinfra.context import ctx_config, ctx_inventory, ctx_state
from pyinfra.operations import server

from .exceptions import (
    CliError,
    UnexpectedExternalError,
    UnexpectedInternalError,
)
from .inventory import make_inventory
from .log import setup_logging
from .prints import (
    print_facts,
    print_inventory,
    print_meta,
    print_results,
    print_state_facts,
    print_state_operations,
    print_support_info,
)
from .util import (
    exec_file,
    get_facts_and_args,
    get_operation_and_args,
    load_deploy_file,
    parse_cli_arg,
)
from .virtualenv import init_virtualenv


def _exit():
    if ctx_state.isset() and state.failed_hosts:
        sys.exit(1)
    sys.exit(0)


def _print_support(ctx, param, value):
    if not value:
        return

    click.echo('--> Support information:', err=True)
    print_support_info()
    ctx.exit()


@click.command()
@click.argument('inventory', nargs=1)
@click.argument('operations', nargs=-1, required=True)
@click.option(
    'verbosity', '-v',
    count=True,
    help='Print meta (-v), input (-vv) and output (-vvv).',
)
@click.option(
    '--dry', is_flag=True, default=False,
    help="Don't execute operations on the target hosts.",
)
@click.option(
    '--limit',
    help='Restrict the target hosts by name and group name.',
    multiple=True,
)
@click.option('--fail-percent', type=int, help='% of hosts allowed to fail.')
@click.option(
    '--data',
    multiple=True,
    help='Override data values, format key=value.',
)
@click.option(
    '--group-data',
    multiple=True,
    help='Paths to load additional group data from (overrides matching keys).',
)
@click.option(
    '--config',
    'config_filename',
    help='Specify config file to use (default: config.py).',
    default='config.py',
)
@click.option(
    '--chdir',
    help='Set the working directory before executing.',
)
# Auth args
@click.option(
    '--sudo', is_flag=True, default=False,
    help='Whether to execute operations with sudo.',
)
@click.option('--sudo-user', help='Which user to sudo when sudoing.')
@click.option(
    '--use-sudo-password', is_flag=True, default=False,
    help='Whether to use a password with sudo.',
)
@click.option('--su-user', help='Which user to su to.')
@click.option('--shell-executable', help='Shell to use (ex: "sh", "cmd", "ps").')
# Operation flow args
@click.option('--parallel', type=int, help='Number of operations to run in parallel.')
@click.option(
    '--no-wait', is_flag=True, default=False,
    help="Don't wait between operations for hosts.",
)
@click.option(
    '--serial', is_flag=True, default=False,
    help='Run operations in serial, host by host.',
)
# SSH connector args
# TODO: remove the non-ssh-prefixed variants
@click.option('--ssh-user', '--user', 'ssh_user', help='SSH user to connect as.')
@click.option('--ssh-port', '--port', 'ssh_port', type=int, help='SSH port to connect to.')
@click.option('--ssh-key', '--key', 'ssh_key', type=click.Path(), help='SSH Private key filename.')
@click.option(
    '--ssh-key-password', '--key-password', 'ssh_key_password',
    help='SSH Private key password.',
)
@click.option('--ssh-password', '--password', 'ssh_password', help='SSH password.')
# WinRM connector args
@click.option('--winrm-username', help='WINRM user to connect as.')
@click.option('--winrm-password', help='WINRM password.')
@click.option('--winrm-port', help='WINRM port to connect to.')
@click.option('--winrm-transport', help='WINRM transport for use.')
# Eager commands (pyinfra --support)
@click.option(
    '--support', is_flag=True, is_eager=True, callback=_print_support,
    help='Print useful information for support and exit.',
)
# Debug args
@click.option(
    '--quiet', is_flag=True, default=False,
    help='Hide most pyinfra output.',
)
@click.option(
    '--debug', is_flag=True, default=False,
    help='Print debug info.',
)
@click.option(
    '--debug-facts', is_flag=True, default=False,
    help='Print facts after generating operations and exit.',
)
@click.option(
    '--debug-operations', is_flag=True, default=False,
    help='Print operations after generating and exit.',
)
@click.version_option(
    version=__version__,
    prog_name='pyinfra',
    message='%(prog)s: v%(version)s',
)
def cli(*args, **kwargs):
    '''
    pyinfra manages the state of one or more servers. It can be used for
    app/service deployment, config management and ad-hoc command execution.

    Documentation: pyinfra.readthedocs.io

    # INVENTORY

    \b
    + a file (inventory.py)
    + hostname (host.net)
    + Comma separated hostnames:
      host-1.net,host-2.net,@local

    # OPERATIONS

    \b
    # Run one or more deploys against the inventory
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...

    \b
    # Run a single operation against the inventory
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra

    \b
    # Execute an arbitrary command against the inventory
    pyinfra INVENTORY exec -- echo "hello world"

    \b
    # Run one or more facts against the inventory
    pyinfra INVENTORY fact server.LinuxName [server.Users]...
    pyinfra INVENTORY fact files.File path=/path/to/file...

    \b
    # Debug the inventory hosts and data
    pyinfra INVENTORY debug-inventory
    '''

    try:
        _main(*args, **kwargs)

    except PyinfraError as e:
        # Re-raise any internal exceptions that aren't handled by click as
        # CliErrors which are.
        if not isinstance(e, click.ClickException):
            message = getattr(e, 'message', e.args[0])
            raise CliError(message)

        raise

    except UnexpectedExternalError:
        # Pass unexpected external exceptions through as-is
        raise

    except Exception as e:
        # Re-raise any unexpected internal exceptions as UnexpectedInternalError
        raise UnexpectedInternalError(e)

    finally:
        if ctx_state.isset() and state.initialised:
            # Triggers any executor disconnect requirements
            disconnect_all(state)


def _main(
    inventory, operations, verbosity, chdir,
    ssh_user, ssh_port, ssh_key, ssh_key_password, ssh_password,
    winrm_username, winrm_password, winrm_port,
    winrm_transport, shell_executable,
    sudo, sudo_user, use_sudo_password, su_user,
    parallel, fail_percent, data, group_data, config_filename,
    dry, limit, no_wait, serial, quiet,
    debug, debug_facts, debug_operations,
    support=None,
):
    # Setup working directory
    #

    if chdir:
        os_chdir(chdir)

    # Setup logging
    #

    if not debug and not sys.warnoptions:
        warnings.simplefilter('ignore')

    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    elif quiet:
        log_level = logging.WARNING

    setup_logging(log_level)

    # Bootstrap any virtualenv
    init_virtualenv()

    #  Check operations are valid and setup command
    #

    # Make a copy before we overwrite
    original_operations = operations

    # Debug (print) inventory + group data
    if operations[0] == 'debug-inventory':
        command = 'debug-inventory'

    # Get one or more facts
    elif operations[0] == 'fact':
        command = 'fact'
        operations = get_facts_and_args(operations[1:])

    # Execute a raw command with server.shell
    elif operations[0] == 'exec':
        command = 'exec'
        operations = operations[1:]

    # Execute one or more deploy files
    elif all(cmd.endswith('.py') for cmd in operations):
        command = 'deploy'

        filenames = []

        for filename in operations[0:]:
            if path.exists(filename):
                filenames.append(filename)
                continue
            if chdir and filename.startswith(chdir):
                correct_filename = path.relpath(filename, chdir)
                logger.warning((
                    'Fixing deploy filename under `--chdir` argument: '
                    f'{filename} -> {correct_filename}'
                ))
                filenames.append(correct_filename)
                continue
            raise CliError('No deploy file: {0}'.format(
                path.join(chdir, filename) if chdir else filename,
            ))

        operations = filenames

    # Operation w/optional args (<module>.<op> ARG1 ARG2 ...)
    elif len(operations[0].split('.')) == 2:
        command = 'op'
        operations = get_operation_and_args(operations)

    else:
        raise CliError('''Invalid operations: {0}

    Operation usage:
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra
    pyinfra INVENTORY exec -- echo "hello world"
    pyinfra INVENTORY fact os [users]...'''.format(operations))

    # Setup state, config & inventory
    #

    cwd = getcwd()
    if cwd not in sys.path:  # ensure cwd is present in sys.path
        sys.path.append(cwd)

    state = State()
    state.cwd = cwd
    ctx_state.set(state)

    if verbosity > 0:
        state.print_fact_info = True
        state.print_noop_info = True

    if verbosity > 1:
        state.print_input = state.print_fact_input = True

    if verbosity > 2:
        state.print_output = state.print_fact_output = True

    if not quiet:
        click.echo('--> Loading config...', err=True)

    config = Config()
    ctx_config.set(config)

    # Load up any config.py from the filesystem
    config_filename = path.join(state.cwd, config_filename)
    if path.exists(config_filename):
        exec_file(config_filename)

    # Lock the current config, this allows us to restore this version after
    # executing deploy files that may alter them.
    config.lock_current_sate()

    # Arg based config overrides
    if sudo:
        config.SUDO = True
        if sudo_user:
            config.SUDO_USER = sudo_user

    if use_sudo_password:
        config.USE_SUDO_PASSWORD = use_sudo_password

    if su_user:
        config.SU_USER = su_user

    if parallel:
        config.PARALLEL = parallel

    if shell_executable:
        config.SHELL = None if shell_executable in ('None', 'null') else shell_executable

    if fail_percent is not None:
        config.FAIL_PERCENT = fail_percent

    if not quiet:
        click.echo('--> Loading inventory...', err=True)

    override_data = {}

    for arg in data:
        key, value = arg.split('=', 1)
        override_data[key] = value

    override_data = {
        key: parse_cli_arg(value)
        for key, value in override_data.items()
    }

    for key, value in (
        ('ssh_user', ssh_user),
        ('ssh_key', ssh_key),
        ('ssh_key_password', ssh_key_password),
        ('ssh_port', ssh_port),
        ('ssh_password', ssh_password),
        ('winrm_username', winrm_username),
        ('winrm_password', winrm_password),
        ('winrm_port', winrm_port),
        ('winrm_transport', winrm_transport),
    ):
        if value:
            override_data[key] = value

    # Load up the inventory from the filesystem
    inventory, inventory_group = make_inventory(
        inventory,
        cwd=state.cwd,
        override_data=override_data,
        group_data_directories=group_data,
    )
    ctx_inventory.set(inventory)

    # Now that we have inventory, apply --limit config override
    initial_limit = None
    if limit:
        all_limit_hosts = []

        for limiter in limit:
            try:
                limit_hosts = inventory.get_group(limiter)
            except NoGroupError:
                limit_hosts = [host for host in inventory if fnmatch(host.name, limiter)]

            if not limit_hosts:
                logger.warning('No host matches found for --limit pattern: {0}'.format(limiter))

            all_limit_hosts.extend(limit_hosts)
        initial_limit = list(set(all_limit_hosts))

    # Initialise the state
    state.init(inventory, config, initial_limit=initial_limit)

    if command == 'debug-inventory':
        print_inventory(state)
        _exit()

    # Connect to the hosts & start handling the user commands
    #

    if not quiet:
        click.echo(err=True)
        click.echo('--> Connecting to hosts...', err=True)

    connect_all(state)

    if command == 'fact':
        if not quiet:
            click.echo(err=True)
            click.echo('--> Gathering facts...', err=True)

        state.print_fact_info = True
        fact_data = {}

        for i, command in enumerate(operations):
            fact_cls, args, kwargs = command
            fact_key = fact_cls.name

            if args or kwargs:
                fact_key = '{0}{1}{2}'.format(
                    fact_cls.name,
                    args or '',
                    ' ({0})'.format(get_kwargs_str(kwargs)) if kwargs else '',
                )

            try:
                fact_data[fact_key] = get_facts(
                    state,
                    fact_cls,
                    args=args,
                    kwargs=kwargs,
                    apply_failed_hosts=False,
                )
            except PyinfraError:
                pass

        print_facts(fact_data)
        _exit()

    if command == 'exec':
        state.print_output = True
        add_op(
            state,
            server.shell,
            ' '.join(operations),
            _allow_cli_mode=True,
        )

    elif command == 'deploy':
        if not quiet:
            click.echo(err=True)
            click.echo('--> Preparing operations...', err=True)

        # Number of "steps" to make = number of files * number of hosts
        for i, filename in enumerate(operations):
            logger.info('Loading: {0}'.format(click.style(filename, bold=True)))
            load_deploy_file(state, filename)
            # Remove any config changes introduced by the deploy file & any includes
            config.reset_locked_state()

    elif command == 'op':
        if not quiet:
            click.echo(err=True)
            click.echo('--> Preparing operation...', err=True)

        op, args = operations
        args, kwargs = args
        kwargs['_allow_cli_mode'] = True

        def print_host_ready(host):
            logger.info('{0}{1} {2}'.format(
                host.print_prefix,
                click.style('Ready:', 'green'),
                click.style(original_operations[0], bold=True),
            ))

        kwargs['_after_host_callback'] = print_host_ready

        add_op(state, op, *args, **kwargs)

    # Print proposed changes, execute unless --dry, and exit
    #

    if not quiet:
        click.echo(err=True)
        click.echo('--> Proposed changes:', err=True)
    print_meta(state)

    # If --debug-facts or --debug-operations, print and exit
    if debug_facts or debug_operations:
        if debug_facts:
            print_state_facts(state)

        if debug_operations:
            print_state_operations(state)

        _exit()

    if dry:
        _exit()

    if not quiet:
        click.echo(err=True)

    if not quiet:
        click.echo('--> Beginning operation run...', err=True)
    run_ops(state, serial=serial, no_wait=no_wait)

    if not quiet:
        click.echo('--> Results:', err=True)
    print_results(state)

    _exit()



# encoding: utf8

from pyinfra.api import FactBase

from .util.packaging import parse_packages

NPM_REGEX = r'^[└├]\─\─\s([a-zA-Z0-9\-]+)@([0-9\.]+)$'


class NpmPackages(FactBase):
    '''
    Returns a dict of installed npm packages globally or in a given directory:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    default = dict

    requires_command = 'npm'

    def command(self, directory=None):
        if directory:
            return (
                'cd {0} && npm list -g --depth=0'
            ).format(directory)
        else:
            return 'npm list -g --depth=0'

    def process(self, output):
        return parse_packages(NPM_REGEX, output)



from pyinfra import logger

from .util import get_call_location, memoize


auth_kwargs = {
    '_sudo': {
        'description': 'Execute/apply any changes with ``sudo``.',
        'default': lambda config: config.SUDO,
    },
    '_sudo_user': {
        'description': 'Execute/apply any changes with ``sudo`` as a non-root user.',
        'default': lambda config: config.SUDO_USER,
    },
    '_use_sudo_login': {
        'description': 'Execute ``sudo`` with a login shell.',
        'default': lambda config: config.USE_SUDO_LOGIN,
    },
    '_use_sudo_password': {
        'description': 'Whether to use a password with ``sudo`` (will ask).',
        'default': lambda config: config.USE_SUDO_PASSWORD,
    },
    '_preserve_sudo_env': {
        'description': 'Preserve the shell environment when using ``sudo``.',
        'default': lambda config: config.PRESERVE_SUDO_ENV,
    },
    '_su_user': {
        'description': 'Execute/apply any changes with ``su``.',
        'default': lambda config: config.SU_USER,
    },
    '_use_su_login': {
        'description': 'Execute ``su`` with a login shell.',
        'default': lambda config: config.USE_SU_LOGIN,
    },
    '_preserve_su_env': {
        'description': 'Preserve the shell environment when using ``su``.',
        'default': lambda config: config.PRESERVE_SU_ENV,
    },
    '_su_shell': {
        'description': (
            'Use this shell (instead of user login shell) when using ``su``). '
            'Only available under Linux, for use when using `su` with a user that '
            'has nologin/similar as their login shell.'
        ),
        'default': lambda config: config.SU_SHELL,
    },
    '_doas': {
        'description': 'Execute/apply any changes with ``doas``.',
        'defailt': lambda config: config.DOAS,
    },
    '_doas_user': {
        'description': 'Execute/apply any changes with ``doas`` as a non-root user.',
        'default': lambda config: config.DOAS_USER,
    },
}


def generate_env(config, value):
    env = config.ENV.copy()

    # TODO: this is to protect against host.data.env being a string or similar,
    # the introduction of using host.data.X for operation kwargs combined with
    # `env` being a commonly defined data variable causes issues.
    # The real fix here is the prefixed `_env` argument.
    if value and isinstance(value, dict):
        env.update(value)

    return env


shell_kwargs = {
    '_shell_executable': {
        'description': 'The shell to use. Defaults to ``sh`` (Unix) or ``cmd`` (Windows).',
        'default': lambda config: config.SHELL,
    },
    '_chdir': {
        'description': 'Directory to switch to before executing the command.',
    },
    '_env': {
        'description': 'Dictionary of environment variables to set.',
        'handler': generate_env,
    },
    '_success_exit_codes': {
        'description': 'List of exit codes to consider a success.',
        'default': lambda config: [0],
    },
    '_timeout': 'Timeout for *each* command executed during the operation.',
    '_get_pty': 'Whether to get a pseudoTTY when executing any commands.',
    '_stdin': 'String or buffer to send to the stdin of any commands.',
}

meta_kwargs = {
    # NOTE: name is the only non-_-prefixed argument
    'name': {
        'description': 'Name of the operation.',
    },
    '_ignore_errors': {
        'description': 'Ignore errors when executing the operation.',
        'default': lambda config: config.IGNORE_ERRORS,
    },
    '_precondition': 'Command to execute & check before the operation commands begin.',
    '_postcondition': 'Command to execute & check after the operation commands complete.',
    '_on_success': 'Callback function to execute on success.',
    '_on_error': 'Callback function to execute on error.',
}

# Execution kwargs are global - ie must be identical for every host
execution_kwargs = {
    '_parallel': {
        'description': 'Run this operation in batches of hosts.',
        'default': lambda config: config.PARALLEL,
    },
    '_run_once': {
        'description': 'Only execute this operation once, on the first host to see it.',
        'default': lambda config: False,
    },
    '_serial': {
        'description': 'Run this operation host by host, rather than in parallel.',
        'default': lambda config: False,
    },
}

OPERATION_KWARGS = {
    'Privilege & user escalation': auth_kwargs,
    'Shell control & features': shell_kwargs,
    'Operation meta & callbacks (not available in facts)': meta_kwargs,
    'Execution strategy (not available in facts, must be the same for all hosts)': execution_kwargs,
}


def _get_internal_key(key):
    if key.startswith('_'):
        return key[1:]
    return key


@memoize
def get_execution_kwarg_keys():
    return [_get_internal_key(key) for key in execution_kwargs.keys()]


@memoize
def get_executor_kwarg_keys():
    keys = set()
    keys.update(auth_kwargs.keys(), shell_kwargs.keys())
    return [_get_internal_key(key) for key in keys]


@memoize
def show_legacy_argument_warning(key, call_location):
    logger.warning((
        '{0}:\n\tGlobal arguments should be prefixed "_", '
        'please us the `{1}` keyword argument in place of `{2}`.'
    ).format(call_location, '_{0}'.format(key), key))


@memoize
def show_legacy_argument_host_data_warning(key):
    logger.warning((
        'Global arguments should be prefixed "_", '
        'please us the `host.data._{0}` keyword argument in place of `host.data.{0}`.'
    ).format(key))


def pop_global_arguments(state, host, kwargs, keys_to_check=None):
    '''
    Pop and return operation global keyword arguments, in preferred order:

    + From the current context (a direct @operator or @deploy function being called)
    + From any current @deploy context (deploy kwargs)
    + From the host data variables
    + From the config variables

    Note this function is only called directly in the @operation & @deploy decorator
    wrappers which the user should pass global arguments prefixed "_". This is to
    avoid any clashes with operation and deploy functions both internal and third
    party.

    This is a bit strange because internally pyinfra uses non-_-prefixed arguments,
    and this function is responsible for the translation between the two.

    TODO: is this wird-ness acceptable? Is it worth updating internal use to _prefix?
    '''

    meta_kwargs = host.current_deploy_kwargs or {}

    global_kwargs = {}
    found_keys = []

    for _, kwarg_configs in OPERATION_KWARGS.items():
        for key, config in kwarg_configs.items():
            internal_key = _get_internal_key(key)

            if keys_to_check and internal_key not in keys_to_check:
                continue

            handler = None
            default = None

            if isinstance(config, dict):
                handler = config.get('handler')
                default = config.get('default')
                if default:
                    default = default(state.config)

            host_default = getattr(host.data, key, None)

            # TODO: remove this additional check in v3
            if host_default is None and internal_key != key:
                host_default = getattr(host.data, internal_key, None)
                if host_default is not None:
                    show_legacy_argument_host_data_warning(internal_key)

            if host_default is not None:
                default = host_default

            if key in kwargs:
                found_keys.append(internal_key)
                value = kwargs.pop(key)

            # TODO: remove this additional check in v3
            elif internal_key in kwargs:
                show_legacy_argument_warning(internal_key, get_call_location(frame_offset=2))
                found_keys.append(internal_key)
                value = kwargs.pop(internal_key)

            else:
                value = meta_kwargs.get(internal_key, default)

            if handler:
                value = handler(state.config, value)

            global_kwargs[internal_key] = value
    return global_kwargs, found_keys



from os import path
from random import shuffle

from pyinfra import inventory, state
from pyinfra.context import ctx_state

from .util import run_cli
from ..paramiko_util import PatchSSHTestCase


class TestCliDeploy(PatchSSHTestCase):
    def test_interdependent_deploy(self):
        ctx_state.reset()

        result = run_cli(
            'somehost',
            path.join('tests', 'deploy', 'deploy_interdependent.py'),
            f'--chdir={path.join("tests", "deploy")}',
        )
        assert result.exit_code == 0, result.stdout

        # Check every operation had commands/changes - this ensures that each
        # combo (add/remove/add) always had changes.
        for host, ops in state.ops.items():
            for _, op in ops.items():
                assert len(op['commands']) > 0


class TestCliDeployState(PatchSSHTestCase):
    def _assert_op_data(self, correct_op_name_and_host_names):
        op_order = state.get_op_order()

        assert (
            len(correct_op_name_and_host_names) == len(op_order)
        ), 'Incorrect number of operations detected'

        for i, (correct_op_name, correct_host_names) in enumerate(
            correct_op_name_and_host_names,
        ):
            op_hash = op_order[i]
            op_meta = state.op_meta[op_hash]

            assert list(op_meta['names'])[0] == correct_op_name

            for host in state.inventory:
                op_hashes = state.meta[host]['op_hashes']
                if correct_host_names is True or host.name in correct_host_names:
                    self.assertIn(op_hash, op_hashes)
                else:
                    self.assertNotIn(op_hash, op_hashes)

    def test_deploy(self):
        task_file_path = path.join('tasks', 'a_task.py')
        nested_task_path = path.join('tasks', 'another_task.py')
        correct_op_name_and_host_names = [
            ('First main operation', True),  # true for all hosts
            ('Second main operation', ('somehost',)),
            ('{0} | First task operation'.format(task_file_path), ('anotherhost',)),
            ('{0} | Task order loop 1'.format(task_file_path), ('anotherhost',)),
            ('{0} | 2nd Task order loop 1'.format(task_file_path), ('anotherhost',)),
            ('{0} | Task order loop 2'.format(task_file_path), ('anotherhost',)),
            ('{0} | 2nd Task order loop 2'.format(task_file_path), ('anotherhost',)),
            (
                '{0} | {1} | Second task operation'.format(task_file_path, nested_task_path),
                ('anotherhost',),
            ),
            ('{0} | First task operation'.format(task_file_path), True),
            ('{0} | Task order loop 1'.format(task_file_path), True),
            ('{0} | 2nd Task order loop 1'.format(task_file_path), True),
            ('{0} | Task order loop 2'.format(task_file_path), True),
            ('{0} | 2nd Task order loop 2'.format(task_file_path), True),
            ('{0} | {1} | Second task operation'.format(task_file_path, nested_task_path), True),
            ('My deploy | First deploy operation', True),
            ('My deploy | My nested deploy | First nested deploy operation', True),
            ('My deploy | Second deploy operation', True),
            ('Loop-0 main operation', True),
            ('Loop-1 main operation', True),
            ('Third main operation', True),
            ('Order loop 1', True),
            ('2nd Order loop 1', True),
            ('Order loop 2', True),
            ('2nd Order loop 2', True),
            ('Final limited operation', ('somehost',)),
        ]

        # Run 3 iterations of the test - each time shuffling the order of the
        # hosts - ensuring that the ordering has no effect on the operation order.
        for _ in range(3):
            ctx_state.reset()

            hosts = ['somehost', 'anotherhost', 'someotherhost']
            shuffle(hosts)

            result = run_cli(
                ','.join(hosts),
                path.join('tests', 'deploy', 'deploy.py'),
                f'--chdir={path.join("tests", "deploy")}',
            )
            assert result.exit_code == 0, result.stdout

            self._assert_op_data(correct_op_name_and_host_names)

    def test_random_deploy(self):
        correct_op_name_and_host_names = [
            ('First main operation', True),
            ('Second main somehost operation', ('somehost',)),
            ('Second main anotherhost operation', ('anotherhost',)),
            ('Third main operation', True),
        ]

        # Run 3 iterations of the test - each time shuffling the order of the
        # hosts - ensuring that the ordering has no effect on the operation order.
        for _ in range(3):
            ctx_state.reset()

            hosts = ['somehost', 'anotherhost', 'someotherhost']
            shuffle(hosts)

            result = run_cli(
                ','.join(hosts),
                path.join('tests', 'deploy', 'deploy_random.py'),
                f'--chdir={path.join("tests", "deploy")}',
            )
            assert result.exit_code == 0, result.stdout

            self._assert_op_data(correct_op_name_and_host_names)

            for hostname, expected_fact_count in (
                ('somehost', 2),
                ('anotherhost', 0),
                ('someotherhost', 1),
            ):
                host = inventory.get_host(hostname)
                assert len(host.facts) == expected_fact_count



'''
The pyinfra facts API. Facts enable pyinfra to collect remote server state which
is used to "diff" with the desired state, producing the final commands required
for a deploy.
'''

import re

from inspect import getcallargs
from socket import (
    error as socket_error,
    timeout as timeout_error,
)

import click
import gevent

from paramiko import SSHException

from pyinfra import logger
from pyinfra.api import StringCommand
from pyinfra.api.arguments import pop_global_arguments
from pyinfra.api.util import (
    get_kwargs_str,
    log_error_or_warning,
    log_host_command_error,
    make_hash,
    print_host_combined_output,
)
from pyinfra.connectors.util import split_combined_output
from pyinfra.progress import progress_spinner

from .arguments import get_executor_kwarg_keys


SUDO_REGEX = r'^sudo: unknown user:'
SU_REGEXES = (
    r'^su: user .+ does not exist',
    r'^su: unknown login',
)


class FactNameMeta(type):
    def __init__(cls, name, bases, attrs):
        module_name = cls.__module__.replace('pyinfra.facts.', '')
        cls.name = f'{module_name}.{cls.__name__}'


class FactBase(object, metaclass=FactNameMeta):
    abstract = True

    shell_executable = None

    requires_command = None

    @staticmethod
    def default():
        '''
        Set the default attribute to be a type (eg list/dict).
        '''

    @staticmethod
    def process(output):
        return '\n'.join(output)

    def process_pipeline(self, args, output):
        return {
            arg: self.process([output[i]])
            for i, arg in enumerate(args)
        }


class ShortFactBase(object, metaclass=FactNameMeta):
    fact = None


def get_short_facts(state, host, short_fact, **kwargs):
    fact_data = get_fact(state, host, short_fact.fact, **kwargs)
    return short_fact.process_data(fact_data)


def _make_command(command_attribute, host_args):
    if callable(command_attribute):
        host_args.pop('self', None)
        return command_attribute(**host_args)
    return command_attribute


def _get_executor_kwargs(state, host, override_kwargs=None, override_kwarg_keys=None):
    if override_kwargs is None:
        override_kwargs = {}
        override_kwarg_keys = []

    # Apply any current op kwargs that *weren't* found in the overrides
    override_kwargs.update({
        key: value
        for key, value in (host.current_op_global_kwargs or {}).items()
        if key not in override_kwarg_keys
    })

    return {
        key: value
        for key, value in override_kwargs.items()
        if key in get_executor_kwarg_keys()
    }


def get_facts(state, *args, **kwargs):
    greenlet_to_host = {
        state.pool.spawn(get_fact, state, host, *args, **kwargs): host
        for host in state.inventory.iter_active_hosts()
    }

    results = {}

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            results[host] = greenlet.get()
            progress(host)

    return results


def get_fact(
    state,
    host,
    cls,
    args=None,
    kwargs=None,
    ensure_hosts=None,
    apply_failed_hosts=True,
    fact_hash=None,
):
    if issubclass(cls, ShortFactBase):
        return get_short_facts(state, host, cls, args=args, ensure_hosts=ensure_hosts)

    with host.facts_lock:
        if fact_hash and fact_hash in host.facts:
            return host.facts[fact_hash]

        return _get_fact(
            state,
            host,
            cls,
            args,
            kwargs,
            ensure_hosts,
            apply_failed_hosts,
            fact_hash,
        )


def _get_fact(
    state,
    host,
    cls,
    args=None,
    kwargs=None,
    ensure_hosts=None,
    apply_failed_hosts=True,
    fact_hash=None,
):
    fact = cls()
    name = fact.name

    args = args or ()
    kwargs = kwargs or {}

    # Get the defaults *and* overrides by popping from kwargs, executor kwargs passed
    # into get_fact override everything else (applied below).
    override_kwargs, override_kwarg_keys = pop_global_arguments(
        state,
        host,
        kwargs,
        keys_to_check=get_executor_kwarg_keys(),
    )

    executor_kwargs = _get_executor_kwargs(
        state,
        host,
        override_kwargs=override_kwargs,
        override_kwarg_keys=override_kwarg_keys,
    )

    if args or kwargs:
        # Merges args & kwargs into a single kwargs dictionary
        kwargs = getcallargs(fact.command, *args, **kwargs)

    logger.debug('Getting fact: {0} ({1}) (ensure_hosts: {2})'.format(
        name, get_kwargs_str(kwargs), ensure_hosts,
    ))

    ignore_errors = (host.current_op_global_kwargs or {}).get(
        'ignore_errors',
        state.config.IGNORE_ERRORS,
    )

    # Facts can override the shell (winrm powershell vs cmd support)
    if fact.shell_executable:
        executor_kwargs['shell_executable'] = fact.shell_executable

    command = _make_command(fact.command, kwargs)
    requires_command = _make_command(fact.requires_command, kwargs)
    if requires_command:
        command = StringCommand(
            # Command doesn't exist, return 0 *or* run & return fact command
            '!', 'command', '-v', requires_command, '>/dev/null', '||', command,
        )

    status = False
    stdout = []

    try:
        status, combined_output_lines = host.run_shell_command(
            command,
            print_output=state.print_fact_output,
            print_input=state.print_fact_input,
            return_combined_output=True,
            **executor_kwargs,
        )
    except (timeout_error, socket_error, SSHException) as e:
        log_host_command_error(
            host, e,
            timeout=executor_kwargs['timeout'],
        )

    stdout, stderr = split_combined_output(combined_output_lines)

    data = fact.default()

    if status:
        if stdout:
            data = fact.process(stdout)
    elif stderr:
        # If we have error output and that error is sudo or su stating the user
        # does not exist, do not fail but instead return the default fact value.
        # This allows for users that don't currently but may be created during
        # other operations.
        first_line = stderr[0]
        if (executor_kwargs['sudo_user'] and re.match(SUDO_REGEX, first_line)):
            status = True
        if (executor_kwargs['su_user'] and any(
            re.match(regex, first_line) for regex in SU_REGEXES
        )):
            status = True

    if status:
        log_message = '{0}{1}'.format(
            host.print_prefix,
            'Loaded fact {0}{1}'.format(
                click.style(name, bold=True),
                f' ({get_kwargs_str(kwargs)})' if kwargs else '',
            ),
        )
        if state.print_fact_info:
            logger.info(log_message)
        else:
            logger.debug(log_message)
    else:
        if not state.print_fact_output:
            print_host_combined_output(host, combined_output_lines)

        log_error_or_warning(host, ignore_errors, description=(
            'could not load fact: {0} {1}'
        ).format(name, get_kwargs_str(kwargs)))

    # Check we've not failed
    if not status and not ignore_errors and apply_failed_hosts:
        state.fail_hosts({host})

    if fact_hash:
        host.facts[fact_hash] = data
    return data


def get_host_fact(state, host, name, args=None, kwargs=None):
    fact_hash = make_hash((name, kwargs, _get_executor_kwargs(state, host)))
    return get_fact(state, host, name, args=args, kwargs=kwargs, fact_hash=fact_hash)


def create_host_fact(state, host, name, data, args=None, kwargs=None):
    fact_hash = make_hash((name, kwargs, _get_executor_kwargs(state, host)))
    host.facts[fact_hash] = data


def delete_host_fact(state, host, name, args=None, kwargs=None):
    fact_hash = make_hash((name, kwargs, _get_executor_kwargs(state, host)))
    host.facts.pop(fact_hash, None)



from unittest import TestCase

from pyinfra.api.util import (
    format_exception,
    get_caller_frameinfo,
    try_int,
)


class TestApiUtil(TestCase):
    def test_try_int_number(self):
        assert try_int('1') == 1

    def test_try_int_fail(self):
        assert try_int('hello') == 'hello'

    def test_get_caller_frameinfo(self):
        def _get_caller_frameinfo():
            return get_caller_frameinfo()

        frameinfo = _get_caller_frameinfo()
        assert frameinfo.lineno == 21  # called by the line above

    def test_format_exception(self):
        exception = Exception('I am a message', 1)
        assert format_exception(exception) == "Exception('I am a message', 1)"



from unittest import TestCase
from unittest.mock import mock_open, patch

from paramiko import ProxyCommand

from pyinfra.connectors.sshuserclient import SSHClient
from pyinfra.connectors.sshuserclient.client import AskPolicy, get_ssh_config

SSH_CONFIG_DATA = '''
# Comment
Host 127.0.0.1
    IdentityFile /id_rsa
    IdentityFile /id_rsa2
    User testuser
    Port 33
    ProxyCommand ssh thing

Include other_file
'''

SSH_CONFIG_OTHER_FILE = '''
Host 192.168.1.1
    User "otheruser"
    ProxyCommand None
    ForwardAgent yes
    UserKnownHostsFile ~/.ssh/test3
'''

SSH_CONFIG_OTHER_FILE_PROXYJUMP = '''
Host 192.168.1.2
    User "otheruser"
    ProxyJump nottestuser@127.0.0.1
    ForwardAgent yes
'''

BAD_SSH_CONFIG_DATA = '''
&
'''

LOOPING_SSH_CONFIG_DATA = '''
Include other_file
'''


class TestSSHUserConfigMissing(TestCase):
    def setUp(self):
        get_ssh_config.cache = {}

    @patch(
        'pyinfra.connectors.sshuserclient.client.path.exists',
        lambda path: False,
    )
    def test_load_ssh_config_no_exist(self):
        client = SSHClient()

        _, config, forward_agent, missing_host_key_policy, host_keys_file \
            = client.parse_config('127.0.0.1')

        assert config.get('port') == 22


@patch(
    'pyinfra.connectors.sshuserclient.client.path.exists',
    lambda path: True,
)
@patch(
    'pyinfra.connectors.sshuserclient.config.glob.iglob',
    lambda path: ['other_file'],
)
@patch(
    'pyinfra.connectors.sshuserclient.config.path.isfile',
    lambda path: True,
)
@patch(
    'pyinfra.connectors.sshuserclient.config.path.expanduser',
    lambda path: path,
)
@patch(
    'pyinfra.connectors.sshuserclient.config.path.isabs',
    lambda path: True,
)
@patch(
    'paramiko.config.LazyFqdn.__str__',
    lambda self: '',
)
class TestSSHUserConfig(TestCase):
    def setUp(self):
        get_ssh_config.cache = {}

    @patch(
        'pyinfra.connectors.sshuserclient.client.open',
        mock_open(read_data=SSH_CONFIG_DATA),
        create=True,
    )
    @patch(
        'pyinfra.connectors.sshuserclient.config.open',
        mock_open(read_data=SSH_CONFIG_OTHER_FILE),
        create=True,
    )
    def test_load_ssh_config(self):
        client = SSHClient()

        _, config, forward_agent, missing_host_key_policy, host_keys_file \
            = client.parse_config('127.0.0.1')

        assert config.get('key_filename') == ['/id_rsa', '/id_rsa2']
        assert config.get('username') == 'testuser'
        assert config.get('port') == 33
        assert isinstance(config.get('sock'), ProxyCommand)
        assert forward_agent is False
        assert isinstance(missing_host_key_policy, AskPolicy)
        assert host_keys_file == '~/.ssh/known_hosts'  # OpenSSH default

        _, other_config, forward_agent, missing_host_key_policy, host_keys_file \
            = client.parse_config('192.168.1.1')

        assert other_config.get('username') == 'otheruser'
        assert forward_agent is True
        assert isinstance(missing_host_key_policy, AskPolicy)
        assert host_keys_file == '~/.ssh/test3'

    @patch(
        'pyinfra.connectors.sshuserclient.client.open',
        mock_open(read_data=BAD_SSH_CONFIG_DATA),
        create=True,
    )
    def test_invalid_ssh_config(self):
        client = SSHClient()

        with self.assertRaises(Exception) as context:
            client.parse_config('127.0.0.1')

        assert context.exception.args[0] == 'Unparsable line &'

    @patch(
        'pyinfra.connectors.sshuserclient.client.open',
        mock_open(read_data=LOOPING_SSH_CONFIG_DATA),
        create=True,
    )
    @patch(
        'pyinfra.connectors.sshuserclient.config.open',
        mock_open(read_data=LOOPING_SSH_CONFIG_DATA),
        create=True,
    )
    def test_include_loop_ssh_config(self):
        client = SSHClient()

        with self.assertRaises(Exception) as context:
            client.parse_config('127.0.0.1')

        assert context.exception.args[0] == 'Include loop detected in ssh config file: other_file'

    @patch(
        'pyinfra.connectors.sshuserclient.client.open',
        mock_open(read_data=SSH_CONFIG_DATA),
        create=True,
    )
    @patch(
        'pyinfra.connectors.sshuserclient.config.open',
        mock_open(read_data=SSH_CONFIG_OTHER_FILE_PROXYJUMP),
        create=True,
    )
    @patch('pyinfra.connectors.sshuserclient.SSHClient.connect')
    @patch('pyinfra.connectors.sshuserclient.SSHClient.gateway')
    def test_load_ssh_config_proxyjump(self, fake_gateway, fake_ssh_connect):
        client = SSHClient()

        # Load the SSH config with ProxyJump configured
        _, config, forward_agent, _, _ = client.parse_config(
            '192.168.1.2',
            {'port': 1022},
            ssh_config_file='other_file',
        )

        fake_ssh_connect.assert_called_once_with(
            '127.0.0.1',
            _pyinfra_ssh_config_file='other_file',
            port='33',
            sock=None,
            username='nottestuser',
        )
        fake_gateway.assert_called_once_with('192.168.1.2', 1022, '192.168.1.2', 1022)



import json
import os

from datetime import datetime
from io import open
from os import listdir, path
from unittest.mock import patch

from pyinfra.api import Config, Inventory
from pyinfra.api.util import get_kwargs_str

from . import logger


def get_command_string(command):
    value = command.get_raw_value()
    masked_value = command.get_masked_value()
    if value == masked_value:
        return value
    else:
        return [value, masked_value]


def make_inventory(hosts=('somehost', 'anotherhost'), **kwargs):
    override_data = kwargs.pop('override_data', {})
    if 'ssh_user' not in override_data:
        override_data['ssh_user'] = 'vagrant'

    return Inventory(
        (hosts, {}),
        override_data=override_data,
        test_group=([
            'somehost',
        ], {
            'group_data': 'hello world',
        }),
        **kwargs
    )


class FakeState(object):
    active = True
    cwd = '/'
    in_op = True
    in_deploy = True
    pipelining = False
    deploy_name = None
    deploy_kwargs = None

    def __init__(self):
        self.inventory = Inventory(([], {}))
        self.config = Config()

    def get_temp_filename(*args):
        return '_tempfile_'


def parse_fact(fact):
    '''
    Convert JSON types to more complex Python types because JSON is lacking.
    '''

    # Handle datetimes
    if isinstance(fact, str) and fact.startswith('datetime:'):
        return datetime.strptime(fact[9:], '%Y-%m-%dT%H:%M:%S')

    elif isinstance(fact, list):
        # Handle sets
        if fact and fact[0] == 'set:':
            return set(parse_fact(value) for value in fact[1:])

        return [parse_fact(value) for value in fact]

    elif isinstance(fact, dict):
        return {
            key: parse_fact(value)
            for key, value in fact.items()
        }

    return fact


class FakeFact(object):
    def __init__(self, data):
        self.data = parse_fact(data)

    def __iter__(self):
        return iter(self.data)

    def __getattr__(self, key):
        return getattr(self.data, key)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, key):
        return key in self.data

    def __call__(self, *args, **kwargs):
        item = self.data

        for arg in args:
            if arg is None:
                continue

            # Support for non-JSON-able fact arguments by  turning them into JSON!
            if isinstance(arg, list):
                arg = json.dumps(arg)

            item = item.get(arg)

        return item

    def __str__(self):
        return str(self.data)

    def __unicode__(self):
        return self.data

    def __eq__(self, other_thing):
        return self.data == other_thing

    def __ne__(self, other_thing):
        return self.data != other_thing

    def get(self, key, default=None):
        if key in self.data:
            return self.data[key]

        return default


class FakeFacts(object):
    def __init__(self, facts):
        self.facts = {
            key: FakeFact(value)
            for key, value in facts.items()
        }

    def __getattr__(self, key):
        return self.facts.get(key)

    def __setitem__(self, key, value):
        self.facts[key] = value

    def _create(self, key, data=None, args=None):
        self.facts[key][args[0]] = data

    def _delete(self, key, args=None):
        self.facts[key].pop(args[0], None)


# TODO: remove after python2 removal, as only required because of different default ordering in 2/3
def _sort_kwargs_str(string):
    return ', '.join(sorted(string.split(', ')))


class FakeHost(object):
    noop_description = None

    # Current context inside an @operation function
    in_op = True
    current_op_hash = None
    current_op_global_kwargs = None

    # Current context inside a @deploy function
    in_deploy = True
    current_deploy_name = None
    current_deploy_kwargs = None
    current_deploy_data = None
    current_deploy_op_order = None

    def __init__(self, name, facts, data):
        self.name = name
        self.fact = FakeFacts(facts)
        self.data = data
        self.connector_data = {}

    @property
    def print_prefix(self):
        return ''

    def noop(self, description):
        self.noop_description = description

    @staticmethod
    def _get_fact_key(fact_cls):
        return '{0}.{1}'.format(fact_cls.__module__.split('.')[-1], fact_cls.__name__)

    def get_fact(self, fact_cls, **kwargs):
        fact_key = self._get_fact_key(fact_cls)
        fact = getattr(self.fact, fact_key, None)
        if fact is None:
            raise KeyError('Missing test fact data: {0}'.format(fact_key))
        if kwargs:
            fact_ordered_keys = {
                _sort_kwargs_str(key): value
                for key, value in fact.items()
            }
            kwargs_str = _sort_kwargs_str(get_kwargs_str(kwargs))
            if kwargs_str not in fact:
                logger.info('Possible missing fact key: {0}'.format(kwargs_str))
            return fact_ordered_keys.get(kwargs_str)
        return fact

    def create_fact(self, fact_cls, data=None, kwargs=None):
        try:
            fact = self.get_fact(fact_cls)
        except KeyError:
            fact_key = self._get_fact_key(fact_cls)
            fact = self.fact[fact_key] = {}
        if kwargs:
            fact[_sort_kwargs_str(get_kwargs_str(kwargs))] = data
        else:
            fact_key = self._get_fact_key(fact_cls)
            self.fact[fact_key] = data

    def delete_fact(self, fact_cls, kwargs=None):
        try:
            fact = self.get_fact(fact_cls)
        except KeyError:
            return

        ordered_kwargs = _sort_kwargs_str(get_kwargs_str(kwargs))
        for key in fact.keys():
            ordered_key = _sort_kwargs_str(key)
            if ordered_key == ordered_kwargs:
                fact.pop(key)
                break


class FakeFile(object):
    _read = False
    _data = None

    def __init__(self, name, data=None):
        self._name = name
        self._data = data

    def read(self, *args, **kwargs):
        if self._read is False:
            self._read = True

            if self._data:
                return self._data
            else:
                return '_test_data_'

        return ''

    def readlines(self, *args, **kwargs):
        if self._read is False:
            self._read = True

            if self._data:
                return self._data.split()
            else:
                return ['_test_data_']

            return []

    def seek(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass


class patch_files(object):
    def __init__(self, local_files):
        directories, files, files_data = self._parse_local_files(local_files)

        self._files = files
        self._files_data = files_data
        self._directories = directories

    @staticmethod
    def _parse_local_files(local_files, prefix=FakeState.cwd):
        files = []
        files_data = {}
        directories = {}

        prefix = path.normpath(prefix)

        for filename, file_data in local_files.get('files', {}).items():
            filepath = path.join(prefix, filename)
            files.append(filepath)
            files_data[filepath] = file_data

        for dirname, dir_files in local_files.get('dirs', {}).items():
            sub_dirname = path.join(prefix, dirname)
            sub_directories, sub_files, sub_files_data = patch_files._parse_local_files(
                dir_files,
                sub_dirname,
            )

            files.extend(sub_files)
            files_data.update(sub_files_data)

            directories[sub_dirname] = {
                'files': list(dir_files['files'].keys()),
                'dirs': list(dir_files['dirs'].keys()),
            }
            directories.update(sub_directories)

        return directories, files, files_data

    def __enter__(self):
        self.patches = [
            patch('pyinfra.operations.files.os.path.exists', self.exists),
            patch('pyinfra.operations.files.os.path.isfile', self.isfile),
            patch('pyinfra.operations.files.os.path.isdir', self.isdir),
            patch('pyinfra.operations.files.os.walk', self.walk),
            patch('pyinfra.operations.files.os.makedirs', lambda path: True),
            patch('pyinfra.api.util.stat', self.stat),
            # Builtin patches
            patch('pyinfra.operations.files.open', self.get_file, create=True),
            patch('pyinfra.operations.server.open', self.get_file, create=True),
            patch('pyinfra.api.util.open', self.get_file, create=True),
        ]

        for patched in self.patches:
            patched.start()

    def __exit__(self, type_, value, traceback):
        for patched in self.patches:
            patched.stop()

    def get_file(self, filename, *args):
        if self.isfile(filename):
            normalized_path = path.normpath(filename)
            return FakeFile(normalized_path, self._files_data.get(normalized_path))

        raise IOError('Missing FakeFile: {0}'.format(filename))

    def exists(self, filename, *args):
        return self.isfile(filename) or self.isdir(filename)

    def isfile(self, filename, *args):
        normalized_path = path.normpath(filename)
        return normalized_path in self._files

    def isdir(self, dirname, *args):
        normalized_path = path.normpath(dirname)
        return normalized_path in self._directories

    def stat(self, pathname):
        if self.isfile(pathname):
            mode_int = 33188  # 644 file
        elif self.isdir(pathname):
            mode_int = 16877  # 755 directory
        else:
            raise IOError('No such file or directory: {0}'.format(pathname))

        return os.stat_result((mode_int, 0, 0, 0, 0, 0, 0, 0, 0, 0))

    def walk(self, dirname, topdown=True, onerror=None, followlinks=False):
        if not self.isdir(dirname):
            return

        normalized_path = path.normpath(dirname)
        dir_definition = self._directories[normalized_path]
        child_dirs = dir_definition.get('dirs', [])
        child_files = dir_definition.get('files', [])

        yield dirname, child_dirs, child_files

        for child in child_dirs:
            full_child = path.join(dirname, child)
            for recursive_return in self.walk(full_child, topdown, onerror, followlinks):
                yield recursive_return


def create_host(name=None, facts=None, data=None):
    '''
    Creates a FakeHost object with attached fact data.
    '''

    real_facts = {}
    facts = facts or {}
    data = data or {}

    for name, fact_data in facts.items():
        real_facts[name] = fact_data

    return FakeHost(name, facts=real_facts, data=data)


class JsonTest(type):
    def __new__(cls, name, bases, attrs):
        # Get the JSON files
        files = listdir(attrs['jsontest_files'])
        files = [f for f in files if f.endswith('.json')]

        test_prefix = attrs.get('jsontest_prefix', 'test_')

        def gen_test(test_name, filename):
            def test(self):
                test_data = json.loads(open(
                    path.join(attrs['jsontest_files'], filename),
                    encoding='utf-8',
                ).read())
                self.jsontest_function(test_name, test_data)

            return test

        # Loop them and create class methods to call the jsontest_function
        for filename in files:
            test_name = filename[:-5]

            # Attach the method
            method_name = '{0}{1}'.format(test_prefix, test_name)
            attrs[method_name] = gen_test(test_name, filename)

        return type.__new__(cls, name, bases, attrs)



from unittest import TestCase

from pyinfra.api.host import (
    HostData,
)


class TestHostData(TestCase):
    def test_host_data(self):
        data = HostData('somehost', {'hello': 'world'})
        assert data.hello == 'world'
        assert data.get('hello') == 'world'

    def test_host_data_multiple(self):
        data = HostData(
            'somehost',
            {'hello': 'world'},
            {'hello': 'not-world', 'another': 'thing'},
        )
        assert data.hello == 'world'
        assert data.get('hello') == 'world'
        assert data.another == 'thing'

    def test_host_data_override(self):
        data = HostData('somehost', {'hello': 'world'})
        assert data.hello == 'world'

        data.hello = 'override-world'
        assert data.hello == 'override-world'

    def test_host_data_missing(self):
        data = HostData('somehost', {'hello': 'world'})

        with self.assertRaises(AttributeError) as context:
            getattr(data, 'not-a-key')

        assert context.exception.args[0] == 'Host `somehost` has no data `not-a-key`'
        assert data.get('not-a-key') is None



from unittest.mock import MagicMock, patch

from pyinfra.api import Config, State
from pyinfra.api.arguments import get_executor_kwarg_keys, pop_global_arguments
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.facts import get_facts
from pyinfra.facts.server import Command

from ..paramiko_util import (
    PatchSSHTestCase,
)
from ..util import make_inventory


def _get_executor_defaults(state, host):
    global_argument_defaults, _ = pop_global_arguments(state, host, {})
    return {
        key: value
        for key, value in global_argument_defaults.items()
        if key in get_executor_kwarg_keys()
    }


class TestOperationsApi(PatchSSHTestCase):
    def test_get_fact(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(state, Command, ('yes',))

        assert fact_data == {anotherhost: 'some-output'}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **_get_executor_defaults(state, anotherhost)
        )

    def test_get_fact_current_op_global_arguments(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)
        anotherhost.current_op_global_kwargs = {
            'sudo': True,
            'sudo_user': 'someuser',
            'use_sudo_password': True,
            'su_user': 'someuser',
            'timeout': 10,
            'env': {'HELLO': 'WORLD'},
        }

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(state, Command, ('yes',))

        assert fact_data == {anotherhost: 'some-output'}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults.update(anotherhost.current_op_global_kwargs)

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults
        )

    def test_get_fact_error(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = False, MagicMock()

            with self.assertRaises(PyinfraError) as context:
                get_facts(state, Command, ('fail command',))

        assert context.exception.args[0] == 'No hosts remaining!'

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'fail command',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **_get_executor_defaults(state, anotherhost)
        )

    def test_get_fact_error_ignore(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)
        anotherhost.current_op_global_kwargs = {
            'ignore_errors': True,
        }

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = False, MagicMock()
            fact_data = get_facts(state, Command, ('fail command',))

        assert fact_data == {anotherhost: None}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'fail command',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **_get_executor_defaults(state, anotherhost)
        )

    def test_get_fact_executor_override_arguments(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(state, Command, ('yes',), {'_sudo': True})

        assert fact_data == {anotherhost: 'some-output'}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults['sudo'] = True

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults
        )

    def test_get_fact_executor_host_data_arguments(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')
        anotherhost.data._sudo = True

        connect_all(state)

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(state, Command, ('yes',))

        assert fact_data == {anotherhost: 'some-output'}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults['sudo'] = True

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults
        )

    def test_get_fact_executor_mixed_arguments(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')
        anotherhost.data._sudo = True
        anotherhost.data._sudo_user = 'this-should-be-overridden'
        anotherhost.data._su_user = 'this-should-be-overridden'

        anotherhost.current_op_global_kwargs = {
            'su_user': 'override-su-user',
        }

        connect_all(state)

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(
                state,
                Command,
                args=('yes',),
                kwargs={'_sudo_user': 'override-sudo-user'},
            )

        assert fact_data == {anotherhost: 'some-output'}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults['sudo'] = True
        defaults['sudo_user'] = 'override-sudo-user'
        defaults['su_user'] = 'override-su-user'

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults
        )

    def test_get_fact_cached(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        fact_hash = 'a-fact-hash'
        cached_fact = {'this is a cached fact'}
        anotherhost = inventory.get_host('anotherhost')
        anotherhost.facts[fact_hash] = cached_fact

        connect_all(state)

        with patch('pyinfra.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(
                state,
                Command,
                args=('yes',),
                kwargs={'_sudo': True},
                fact_hash=fact_hash,
            )

        assert fact_data == {anotherhost: cached_fact}
        fake_run_command.assert_not_called()
